from flask import Flask

from database import db_session, init_db
from flask_graphql import GraphQL
from schema import schema

app = Flask(__name__)
app.debug = True

default_query = '''
{
  allEmployees {
    edges {
      node {
        id,
        name,
        department {
          id,
          name
        },
        role {
          id,
          name
        }
      }
    }
  }
}'''.strip()

GraphQL(app, schema=schema, default_query=default_query)


@app.teardown_appcontext
def shutdown_session(exception=None):
    db_session.remove()

if __name__ == '__main__':
    init_db()
    app.run()



from contextlib import contextmanager
from functools import partial, reduce


class Plugin(object):

    def contribute_to_schema(self, schema):
        self.schema = schema


def apply_function(a, b):
    return b(a)


class PluginManager(object):

    PLUGIN_FUNCTIONS = ('get_default_namedtype_name', )

    def __init__(self, schema, plugins=[]):
        self.schema = schema
        self.plugins = []
        for plugin in plugins:
            self.add_plugin(plugin)

    def add_plugin(self, plugin):
        if hasattr(plugin, 'contribute_to_schema'):
            plugin.contribute_to_schema(self.schema)
        self.plugins.append(plugin)

    def get_plugin_functions(self, function):
        for plugin in self.plugins:
            if not hasattr(plugin, function):
                continue
            yield getattr(plugin, function)

    def __getattr__(self, name):
        functions = self.get_plugin_functions(name)
        return partial(reduce, apply_function, functions)

    def __contains__(self, name):
        return name in self.PLUGIN_FUNCTIONS

    @contextmanager
    def context_execution(self, **executor):
        contexts = []
        functions = self.get_plugin_functions('context_execution')
        for f in functions:
            context = f(executor)
            executor = context.__enter__()
            contexts.append((context, executor))
        yield executor
        for context, value in contexts[::-1]:
            context.__exit__(None, None, None)



from singledispatch import singledispatch
from sqlalchemy import types
from sqlalchemy.orm import interfaces

from ...core.classtypes.enum import Enum
from ...core.types.scalars import ID, Boolean, Float, Int, String
from .fields import ConnectionOrListField, SQLAlchemyModelField

try:
    from sqlalchemy_utils.types.choice import ChoiceType
except ImportError:
    class ChoiceType(object):
        pass


def convert_sqlalchemy_relationship(relationship):
    direction = relationship.direction
    model = relationship.mapper.entity
    model_field = SQLAlchemyModelField(model, description=relationship.doc)
    if direction == interfaces.MANYTOONE:
        return model_field
    elif (direction == interfaces.ONETOMANY or
          direction == interfaces.MANYTOMANY):
        return ConnectionOrListField(model_field)


def convert_sqlalchemy_column(column):
    return convert_sqlalchemy_type(getattr(column, 'type', None), column)


@singledispatch
def convert_sqlalchemy_type(type, column):
    raise Exception(
        "Don't know how to convert the SQLAlchemy field %s (%s)" % (column, column.__class__))


@convert_sqlalchemy_type.register(types.Date)
@convert_sqlalchemy_type.register(types.DateTime)
@convert_sqlalchemy_type.register(types.Time)
@convert_sqlalchemy_type.register(types.String)
@convert_sqlalchemy_type.register(types.Text)
@convert_sqlalchemy_type.register(types.Unicode)
@convert_sqlalchemy_type.register(types.UnicodeText)
@convert_sqlalchemy_type.register(types.Enum)
def convert_column_to_string(type, column):
    return String(description=column.doc)


@convert_sqlalchemy_type.register(types.SmallInteger)
@convert_sqlalchemy_type.register(types.BigInteger)
@convert_sqlalchemy_type.register(types.Integer)
def convert_column_to_int_or_id(type, column):
    if column.primary_key:
        return ID(description=column.doc)
    else:
        return Int(description=column.doc)


@convert_sqlalchemy_type.register(types.Boolean)
def convert_column_to_boolean(type, column):
    return Boolean(description=column.doc)


@convert_sqlalchemy_type.register(types.Float)
@convert_sqlalchemy_type.register(types.Numeric)
def convert_column_to_float(type, column):
    return Float(description=column.doc)


@convert_sqlalchemy_type.register(ChoiceType)
def convert_column_to_enum(type, column):
    name = '{}_{}'.format(column.table.name, column.name).upper()
    return Enum(name, type.choices, description=column.doc)



from graphene import ObjectType, Schema, String

from ..utils import get_session


def test_get_session():
    session = 'My SQLAlchemy session'
    schema = Schema(session=session)

    class Query(ObjectType):
        x = String()

        def resolve_x(self, args, info):
            return get_session(info)

    query = '''
        query ReporterQuery {
            x
        }
    '''

    schema = Schema(query=Query, session=session)
    result = schema.execute(query)
    assert not result.errors
    assert result.data['x'] == session



import inspect
from collections import OrderedDict

from graphql.core.execution.executor import Executor
from graphql.core.execution.middlewares.sync import \
    SynchronousExecutionMiddleware
from graphql.core.type import GraphQLSchema as _GraphQLSchema
from graphql.core.utils.introspection_query import introspection_query
from graphql.core.utils.schema_printer import print_schema

from graphene import signals

from ..plugins import CamelCase, PluginManager
from .classtypes.base import ClassType
from .types.base import InstanceType


class GraphQLSchema(_GraphQLSchema):

    def __init__(self, schema, *args, **kwargs):
        self.graphene_schema = schema
        super(GraphQLSchema, self).__init__(*args, **kwargs)


class Schema(object):
    _executor = None

    def __init__(self, query=None, mutation=None, subscription=None,
                 name='Schema', executor=None, plugins=None, auto_camelcase=True, **options):
        self._types_names = {}
        self._types = {}
        self.mutation = mutation
        self.query = query
        self.subscription = subscription
        self.name = name
        self.executor = executor
        plugins = plugins or []
        if auto_camelcase:
            plugins.append(CamelCase())
        self.plugins = PluginManager(self, plugins)
        self.options = options
        signals.init_schema.send(self)

    def __repr__(self):
        return '<Schema: %s (%s)>' % (str(self.name), hash(self))

    def __getattr__(self, name):
        if name in self.plugins:
            return getattr(self.plugins, name)
        return super(Schema, self).__getattr__(name)

    def T(self, _type):
        if not _type:
            return
        if isinstance(_type, ClassType):
            _type = type(_type)
        is_classtype = inspect.isclass(_type) and issubclass(_type, ClassType)
        is_instancetype = isinstance(_type, InstanceType)
        if is_classtype or is_instancetype:
            if _type not in self._types:
                internal_type = _type.internal_type(self)
                self._types[_type] = internal_type
                if is_classtype:
                    self.register(_type)
            return self._types[_type]
        else:
            return _type

    @property
    def executor(self):
        if not self._executor:
            self._executor = Executor(
                [SynchronousExecutionMiddleware()], map_type=OrderedDict)
        return self._executor

    @executor.setter
    def executor(self, value):
        self._executor = value

    @property
    def schema(self):
        if not self.query:
            raise Exception('You have to define a base query type')
        return GraphQLSchema(
            self,
            query=self.T(self.query),
            mutation=self.T(self.mutation),
            subscription=self.T(self.subscription))

    def register(self, object_type, force=False):
        type_name = object_type._meta.type_name
        registered_object_type = not force and self._types_names.get(type_name, None)
        if registered_object_type:
            assert registered_object_type == object_type, 'Type {} already registered with other object type'.format(
                type_name)
        self._types_names[object_type._meta.type_name] = object_type
        return object_type

    def objecttype(self, type):
        name = getattr(type, 'name', None)
        if name:
            objecttype = self._types_names.get(name, None)
            if objecttype and inspect.isclass(
                    objecttype) and issubclass(objecttype, ClassType):
                return objecttype

    def __str__(self):
        return print_schema(self.schema)

    def setup(self):
        assert self.query, 'The base query type is not set'
        self.T(self.query)

    def get_type(self, type_name):
        self.setup()
        if type_name not in self._types_names:
            raise KeyError('Type %r not found in %r' % (type_name, self))
        return self._types_names[type_name]

    @property
    def types(self):
        return self._types_names

    def execute(self, request='', root=None, args=None, **kwargs):
        kwargs = dict(kwargs, request=request, root=root, args=args, schema=self.schema)
        with self.plugins.context_execution(**kwargs) as execute_kwargs:
            return self.executor.execute(**execute_kwargs)

    def introspect(self):
        return self.execute(introspection_query).data



from .inputobjecttype import InputObjectType
from .interface import Interface
from .mutation import Mutation
from .objecttype import ObjectType
from .options import Options
from .scalar import Scalar
from .enum import Enum
from .uniontype import UnionType

__all__ = [
    'InputObjectType',
    'Interface',
    'Mutation',
    'ObjectType',
    'Options',
    'Scalar',
    'Enum',
    'UnionType']



from ..fields import DjangoConnectionField
from .utils import get_filtering_args_from_filterset, get_filterset_class


class DjangoFilterConnectionField(DjangoConnectionField):

    def __init__(self, type, fields=None, order_by=None,
                 extra_filter_meta=None, filterset_class=None,
                 *args, **kwargs):

        self.order_by = order_by or type._meta.filter_order_by
        self.fields = fields or type._meta.filter_fields
        meta = dict(model=type._meta.model,
                    fields=self.fields,
                    order_by=self.order_by)
        if extra_filter_meta:
            meta.update(extra_filter_meta)
        self.filterset_class = get_filterset_class(filterset_class, **meta)
        self.filtering_args = get_filtering_args_from_filterset(self.filterset_class, type)
        kwargs.setdefault('args', {})
        kwargs['args'].update(**self.filtering_args)
        super(DjangoFilterConnectionField, self).__init__(type, *args, **kwargs)

    def get_queryset(self, qs, args, info):
        filterset_class = self.filterset_class
        filter_kwargs = self.get_filter_kwargs(args)
        order = self.get_order(args)
        if order:
            qs = qs.order_by(order)
        return filterset_class(data=filter_kwargs, queryset=qs)

    def get_filter_kwargs(self, args):
        return {k: v for k, v in args.items() if k in self.filtering_args}

    def get_order(self, args):
        return args.get('order_by', None)



from graphql.core.type import (GraphQLBoolean, GraphQLFloat, GraphQLID,
                               GraphQLInt, GraphQLString)

from .base import MountedType


class ScalarType(MountedType):

    def internal_type(self, schema):
        return self._internal_type


class String(ScalarType):
    _internal_type = GraphQLString


class Int(ScalarType):
    _internal_type = GraphQLInt


class Boolean(ScalarType):
    _internal_type = GraphQLBoolean


class ID(ScalarType):
    _internal_type = GraphQLID


class Float(ScalarType):
    _internal_type = GraphQLFloat



from sqlalchemy.ext.declarative.api import DeclarativeMeta
from sqlalchemy.orm.query import Query

from graphene.utils import LazyList


def get_type_for_model(schema, model):
    schema = schema
    types = schema.types.values()
    for _type in types:
        type_model = hasattr(_type, '_meta') and getattr(
            _type._meta, 'model', None)
        if model == type_model:
            return _type


def get_session(info):
    schema = info.schema.graphene_schema
    return schema.options.get('session')


def get_query(model, info):
    query = getattr(model, 'query', None)
    if not query:
        session = get_session(info)
        if not session:
            raise Exception('A query in the model Base or a session in the schema is required for querying.\n'
                            'Read more http://graphene-python.org/docs/sqlalchemy/tips/#querying')
        query = session.query(model)
    return query


class WrappedQuery(LazyList):

    def __len__(self):
        # Dont calculate the length using len(query), as this will
        # evaluate the whole queryset and return it's length.
        # Use .count() instead
        return self._origin.count()


def maybe_query(value):
    if isinstance(value, Query):
        return WrappedQuery(value)
    return value


def is_mapped(obj):
    return isinstance(obj, DeclarativeMeta)



import six
from graphql.core.type import GraphQLEnumType, GraphQLEnumValue

from .base import ClassTypeMeta, ClassType
from ..types.base import MountedType
from ...utils.enum import Enum as PyEnum


class EnumMeta(ClassTypeMeta):

    def construct(cls, bases, attrs):
        __enum__ = attrs.get('__enum__', None)
        if not cls._meta.abstract and not __enum__:
            __enum__ = PyEnum(cls._meta.type_name, attrs)
            setattr(cls, '__enum__', __enum__)
        if __enum__:
            for k, v in __enum__.__members__.items():
                attrs[k] = v.value
        return super(EnumMeta, cls).construct(bases, attrs)

    def __call__(cls, *args, **kwargs):
        if cls is Enum:
            return cls.create_enum(*args, **kwargs)
        return super(EnumMeta, cls).__call__(*args, **kwargs)

    def create_enum(cls, name, names=None, description=None):
        attrs = {
            '__enum__': PyEnum(name, names)
        }
        if description:
            attrs['__doc__'] = description
        return type(name, (Enum,), attrs)


class Enum(six.with_metaclass(EnumMeta, ClassType, MountedType)):

    class Meta:
        abstract = True

    @classmethod
    def internal_type(cls, schema):
        if cls._meta.abstract:
            raise Exception("Abstract Enum don't have a specific type.")

        values = {k: GraphQLEnumValue(v.value) for k, v in cls.__enum__.__members__.items()}
        # GraphQLEnumValue
        return GraphQLEnumType(
            cls._meta.type_name,
            values=values,
            description=cls._meta.description,
        )



from django.db import models

from ...core.classtypes.enum import Enum
from ...core.types.custom_scalars import DateTime, JSONString
from ...core.types.definitions import List
from ...core.types.scalars import ID, Boolean, Float, Int, String
from ...utils import to_const
from .compat import (ArrayField, HStoreField, JSONField, RangeField,
                     RelatedObject, UUIDField)
from .utils import get_related_model, import_single_dispatch

singledispatch = import_single_dispatch()


def convert_choices(choices):
    for value, name in choices:
        yield to_const(name), value


def convert_django_field_with_choices(field):
    choices = getattr(field, 'choices', None)
    if choices:
        meta = field.model._meta
        name = '{}_{}_{}'.format(meta.app_label, meta.object_name, field.name)
        return Enum(name.upper(), list(convert_choices(choices)), description=field.help_text)
    return convert_django_field(field)


@singledispatch
def convert_django_field(field):
    raise Exception(
        "Don't know how to convert the Django field %s (%s)" %
        (field, field.__class__))


@convert_django_field.register(models.CharField)
@convert_django_field.register(models.TextField)
@convert_django_field.register(models.EmailField)
@convert_django_field.register(models.SlugField)
@convert_django_field.register(models.URLField)
@convert_django_field.register(models.GenericIPAddressField)
@convert_django_field.register(models.FileField)
@convert_django_field.register(UUIDField)
def convert_field_to_string(field):
    return String(description=field.help_text)


@convert_django_field.register(models.AutoField)
def convert_field_to_id(field):
    return ID(description=field.help_text)


@convert_django_field.register(models.PositiveIntegerField)
@convert_django_field.register(models.PositiveSmallIntegerField)
@convert_django_field.register(models.SmallIntegerField)
@convert_django_field.register(models.BigIntegerField)
@convert_django_field.register(models.IntegerField)
def convert_field_to_int(field):
    return Int(description=field.help_text)


@convert_django_field.register(models.BooleanField)
def convert_field_to_boolean(field):
    return Boolean(description=field.help_text, required=True)


@convert_django_field.register(models.NullBooleanField)
def convert_field_to_nullboolean(field):
    return Boolean(description=field.help_text)


@convert_django_field.register(models.DecimalField)
@convert_django_field.register(models.FloatField)
def convert_field_to_float(field):
    return Float(description=field.help_text)


@convert_django_field.register(models.DateField)
def convert_date_to_string(field):
    return DateTime(description=field.help_text)


@convert_django_field.register(models.ManyToManyField)
@convert_django_field.register(models.ManyToOneRel)
@convert_django_field.register(models.ManyToManyRel)
def convert_field_to_list_or_connection(field):
    from .fields import DjangoModelField, ConnectionOrListField
    model_field = DjangoModelField(get_related_model(field))
    return ConnectionOrListField(model_field)


# For Django 1.6
@convert_django_field.register(RelatedObject)
def convert_relatedfield_to_djangomodel(field):
    from .fields import DjangoModelField, ConnectionOrListField
    model_field = DjangoModelField(field.model)
    return ConnectionOrListField(model_field)


@convert_django_field.register(models.OneToOneField)
@convert_django_field.register(models.ForeignKey)
def convert_field_to_djangomodel(field):
    from .fields import DjangoModelField
    return DjangoModelField(get_related_model(field), description=field.help_text)


@convert_django_field.register(ArrayField)
def convert_postgres_array_to_list(field):
    base_type = convert_django_field(field.base_field)
    return List(base_type, description=field.help_text)


@convert_django_field.register(HStoreField)
@convert_django_field.register(JSONField)
def convert_posgres_field_to_string(field):
    return JSONString(description=field.help_text)


@convert_django_field.register(RangeField)
def convert_posgres_range_to_string(field):
    inner_type = convert_django_field(field.base_field)
    return List(inner_type, description=field.help_text)



import binascii

from django.core.exceptions import ValidationError
from django.forms import CharField, Field, IntegerField, MultipleChoiceField
from django.utils.translation import ugettext_lazy as _
from graphql_relay import from_global_id


class GlobalIDFormField(Field):
    default_error_messages = {
        'invalid': _('Invalid ID specified.'),
    }

    def clean(self, value):
        if not value and not self.required:
            return None

        try:
            gid = from_global_id(value)
        except (TypeError, ValueError, UnicodeDecodeError, binascii.Error):
            raise ValidationError(self.error_messages['invalid'])

        try:
            IntegerField().clean(gid.id)
            CharField().clean(gid.type)
        except ValidationError:
            raise ValidationError(self.error_messages['invalid'])

        return value


class GlobalIDMultipleChoiceField(MultipleChoiceField):
    default_error_messages = {
        'invalid_choice': _('One of the specified IDs was invalid (%(value)s).'),
        'invalid_list': _('Enter a list of values.'),
    }

    def valid_value(self, value):
        # Clean will raise a validation error if there is a problem
        GlobalIDFormField().clean(value)
        return True



from graphene.contrib.sqlalchemy.types import (
    SQLAlchemyObjectType,
    SQLAlchemyNode
)
from graphene.contrib.sqlalchemy.fields import (
    SQLAlchemyConnectionField,
    SQLAlchemyModelField
)

__all__ = ['SQLAlchemyObjectType', 'SQLAlchemyNode',
           'SQLAlchemyConnectionField', 'SQLAlchemyModelField']



from django.db import models
from django.db.models.manager import Manager
from django.db.models.query import QuerySet

from graphene.utils import LazyList

from .compat import RelatedObject

try:
    import django_filters  # noqa
    DJANGO_FILTER_INSTALLED = True
except (ImportError, AttributeError):
    # AtributeError raised if DjangoFilters installed with a incompatible Django Version
    DJANGO_FILTER_INSTALLED = False


def get_type_for_model(schema, model):
    schema = schema
    types = schema.types.values()
    for _type in types:
        type_model = hasattr(_type, '_meta') and getattr(
            _type._meta, 'model', None)
        if model == type_model:
            return _type


def get_reverse_fields(model):
    for name, attr in model.__dict__.items():
        # Django =>1.9 uses 'rel', django <1.9 uses 'related'
        related = getattr(attr, 'rel', None) or \
            getattr(attr, 'related', None)
        if isinstance(related, RelatedObject):
            # Hack for making it compatible with Django 1.6
            new_related = RelatedObject(related.parent_model, related.model, related.field)
            new_related.name = name
            yield new_related
        elif isinstance(related, models.ManyToOneRel):
            yield related
        elif isinstance(related, models.ManyToManyRel) and not related.symmetrical:
            yield related


class WrappedQueryset(LazyList):

    def __len__(self):
        # Dont calculate the length using len(queryset), as this will
        # evaluate the whole queryset and return it's length.
        # Use .count() instead
        return self._origin.count()


def maybe_queryset(value):
    if isinstance(value, Manager):
        value = value.get_queryset()
    if isinstance(value, QuerySet):
        return WrappedQueryset(value)
    return value


def get_related_model(field):
    if hasattr(field, 'rel'):
        # Django 1.6, 1.7
        return field.rel.to
    return field.related_model


def import_single_dispatch():
    try:
        from functools import singledispatch
    except ImportError:
        singledispatch = None

    if not singledispatch:
        try:
            from singledispatch import singledispatch
        except ImportError:
            pass

    if not singledispatch:
        raise Exception(
            "It seems your python version does not include "
            "functools.singledispatch. Please install the 'singledispatch' "
            "package. More information here: "
            "https://pypi.python.org/pypi/singledispatch"
        )

    return singledispatch



from __future__ import absolute_import

from sqlalchemy import Column, Date, ForeignKey, Integer, String, Table
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()

association_table = Table('association', Base.metadata,
                          Column('pet_id', Integer, ForeignKey('pets.id')),
                          Column('reporter_id', Integer, ForeignKey('reporters.id')))


class Editor(Base):
    __tablename__ = 'editors'
    editor_id = Column(Integer(), primary_key=True)
    name = Column(String(100))


class Pet(Base):
    __tablename__ = 'pets'
    id = Column(Integer(), primary_key=True)
    name = Column(String(30))
    reporter_id = Column(Integer(), ForeignKey('reporters.id'))


class Reporter(Base):
    __tablename__ = 'reporters'
    id = Column(Integer(), primary_key=True)
    first_name = Column(String(30))
    last_name = Column(String(30))
    email = Column(String())
    pets = relationship('Pet', secondary=association_table, backref='reporters')
    articles = relationship('Article', backref='reporter')


class Article(Base):
    __tablename__ = 'articles'
    id = Column(Integer(), primary_key=True)
    headline = Column(String(100))
    pub_date = Column(Date())
    reporter_id = Column(Integer(), ForeignKey('reporters.id'))



from py.test import raises
from sqlalchemy import Column, Table, types
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils.types.choice import ChoiceType

import graphene
from graphene.contrib.sqlalchemy.converter import (convert_sqlalchemy_column,
                                                   convert_sqlalchemy_relationship)
from graphene.contrib.sqlalchemy.fields import (ConnectionOrListField,
                                                SQLAlchemyModelField)

from .models import Article, Pet, Reporter


def assert_column_conversion(sqlalchemy_type, graphene_field, **kwargs):
    column = Column(sqlalchemy_type, doc='Custom Help Text', **kwargs)
    graphene_type = convert_sqlalchemy_column(column)
    assert isinstance(graphene_type, graphene_field)
    field = graphene_type.as_field()
    assert field.description == 'Custom Help Text'
    return field


def test_should_unknown_sqlalchemy_field_raise_exception():
    with raises(Exception) as excinfo:
        convert_sqlalchemy_column(None)
    assert 'Don\'t know how to convert the SQLAlchemy field' in str(excinfo.value)


def test_should_date_convert_string():
    assert_column_conversion(types.Date(), graphene.String)


def test_should_datetime_convert_string():
    assert_column_conversion(types.DateTime(), graphene.String)


def test_should_time_convert_string():
    assert_column_conversion(types.Time(), graphene.String)


def test_should_string_convert_string():
    assert_column_conversion(types.String(), graphene.String)


def test_should_text_convert_string():
    assert_column_conversion(types.Text(), graphene.String)


def test_should_unicode_convert_string():
    assert_column_conversion(types.Unicode(), graphene.String)


def test_should_unicodetext_convert_string():
    assert_column_conversion(types.UnicodeText(), graphene.String)


def test_should_enum_convert_string():
    assert_column_conversion(types.Enum(), graphene.String)


def test_should_small_integer_convert_int():
    assert_column_conversion(types.SmallInteger(), graphene.Int)


def test_should_big_integer_convert_int():
    assert_column_conversion(types.BigInteger(), graphene.Int)


def test_should_integer_convert_int():
    assert_column_conversion(types.Integer(), graphene.Int)


def test_should_integer_convert_id():
    assert_column_conversion(types.Integer(), graphene.ID, primary_key=True)


def test_should_boolean_convert_boolean():
    assert_column_conversion(types.Boolean(), graphene.Boolean)


def test_should_float_convert_float():
    assert_column_conversion(types.Float(), graphene.Float)


def test_should_numeric_convert_float():
    assert_column_conversion(types.Numeric(), graphene.Float)


def test_should_choice_convert_enum():
    TYPES = [
        (u'es', u'Spanish'),
        (u'en', u'English')
    ]
    column = Column(ChoiceType(TYPES), doc='Language', name='language')
    Base = declarative_base()

    Table('translatedmodel', Base.metadata, column)
    graphene_type = convert_sqlalchemy_column(column)
    assert issubclass(graphene_type, graphene.Enum)
    assert graphene_type._meta.type_name == 'TRANSLATEDMODEL_LANGUAGE'
    assert graphene_type._meta.description == 'Language'
    assert graphene_type.__enum__.__members__['es'].value == 'Spanish'
    assert graphene_type.__enum__.__members__['en'].value == 'English'


def test_should_manytomany_convert_connectionorlist():
    graphene_type = convert_sqlalchemy_relationship(Reporter.pets.property)
    assert isinstance(graphene_type, ConnectionOrListField)
    assert isinstance(graphene_type.type, SQLAlchemyModelField)
    assert graphene_type.type.model == Pet


def test_should_manytoone_convert_connectionorlist():
    field = convert_sqlalchemy_relationship(Article.reporter.property)
    assert isinstance(field, SQLAlchemyModelField)
    assert field.model == Reporter


def test_should_onetomany_convert_model():
    graphene_type = convert_sqlalchemy_relationship(Reporter.articles.property)
    assert isinstance(graphene_type, ConnectionOrListField)
    assert isinstance(graphene_type.type, SQLAlchemyModelField)
    assert graphene_type.type.model == Article



import datetime
import json

from graphql.core.language import ast

from ...core.classtypes.scalar import Scalar


class JSONString(Scalar):
    '''JSON String'''

    @staticmethod
    def serialize(dt):
        return json.dumps(dt)

    @staticmethod
    def parse_literal(node):
        if isinstance(node, ast.StringValue):
            return json.dumps(node.value)

    @staticmethod
    def parse_value(value):
        return json.dumps(value)


class DateTime(Scalar):
    '''DateTime in ISO 8601 format'''

    @staticmethod
    def serialize(dt):
        return dt.isoformat()

    @staticmethod
    def parse_literal(node):
        if isinstance(node, ast.StringValue):
            return datetime.datetime.strptime(
                node.value, "%Y-%m-%dT%H:%M:%S.%f")

    @staticmethod
    def parse_value(value):
        return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f")



from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, func
from sqlalchemy.orm import backref, relationship

from database import Base


class Department(Base):
    __tablename__ = 'department'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class Role(Base):
    __tablename__ = 'roles'
    role_id = Column(Integer, primary_key=True)
    name = Column(String)


class Employee(Base):
    __tablename__ = 'employee'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    # Use default=func.now() to set the default hiring time
    # of an Employee to be the current time when an
    # Employee record was created
    hired_on = Column(DateTime, default=func.now())
    department_id = Column(Integer, ForeignKey('department.id'))
    role_id = Column(Integer, ForeignKey('roles.role_id'))
    # Use cascade='delete,all' to propagate the deletion of a Department onto its Employees
    department = relationship(
        Department,
        backref=backref('employees',
                        uselist=True,
                        cascade='delete,all'))
    role = relationship(
        Role,
        backref=backref('roles',
                        uselist=True,
                        cascade='delete,all'))



import importlib
import json
from distutils.version import StrictVersion
from optparse import make_option

from django import get_version as get_django_version
from django.core.management.base import BaseCommand, CommandError

LT_DJANGO_1_8 = StrictVersion(get_django_version()) < StrictVersion('1.8')

if LT_DJANGO_1_8:
    class CommandArguments(BaseCommand):
        option_list = BaseCommand.option_list + (
            make_option(
                '--schema',
                type=str,
                dest='schema',
                default='',
                help='Django app containing schema to dump, e.g. myproject.core.schema',
            ),
            make_option(
                '--out',
                type=str,
                dest='out',
                default='',
                help='Output file (default: schema.json)'
            ),
        )
else:
    class CommandArguments(BaseCommand):

        def add_arguments(self, parser):
            from django.conf import settings
            parser.add_argument(
                '--schema',
                type=str,
                dest='schema',
                default=getattr(settings, 'GRAPHENE_SCHEMA', ''),
                help='Django app containing schema to dump, e.g. myproject.core.schema')

            parser.add_argument(
                '--out',
                type=str,
                dest='out',
                default=getattr(settings, 'GRAPHENE_SCHEMA_OUTPUT', 'schema.json'),
                help='Output file (default: schema.json)')


class Command(CommandArguments):
    help = 'Dump Graphene schema JSON to file'
    can_import_settings = True

    def save_file(self, out, schema_dict):
        with open(out, 'w') as outfile:
            json.dump(schema_dict, outfile)

    def handle(self, *args, **options):
        from django.conf import settings
        schema = options.get('schema') or getattr(settings, 'GRAPHENE_SCHEMA', '')
        out = options.get('out') or getattr(settings, 'GRAPHENE_SCHEMA_OUTPUT', 'schema.json')

        if schema == '':
            raise CommandError('Specify schema on GRAPHENE_SCHEMA setting or by using --schema')
        i = importlib.import_module(schema)

        schema_dict = {'data': i.schema.introspect()}
        self.save_file(out, schema_dict)

        style = getattr(self, 'style', None)
        SUCCESS = getattr(style, 'SUCCESS', lambda x: x)

        self.stdout.write(SUCCESS('Successfully dumped GraphQL schema to %s' % out))



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker

engine = create_engine('sqlite:///database.sqlite3', convert_unicode=True)
db_session = scoped_session(sessionmaker(autocommit=False,
                                         autoflush=False,
                                         bind=engine))
Base = declarative_base()
Base.query = db_session.query_property()


def init_db():
    # import all modules here that might define models so that
    # they will be registered properly on the metadata.  Otherwise
    # you will have to import them first before calling init_db()
    from models import Department, Employee, Role
    Base.metadata.drop_all(bind=engine)
    Base.metadata.create_all(bind=engine)

    # Create the fixtures
    engineering = Department(name='Engineering')
    db_session.add(engineering)
    hr = Department(name='Human Resources')
    db_session.add(hr)

    manager = Role(name='manager')
    db_session.add(manager)
    engineer = Role(name='engineer')
    db_session.add(engineer)

    peter = Employee(name='Peter', department=engineering, role=engineer)
    db_session.add(peter)
    roy = Employee(name='Roy', department=engineering, role=engineer)
    db_session.add(roy)
    tracy = Employee(name='Tracy', department=hr, role=manager)
    db_session.add(tracy)
    db_session.commit()



from graphql.core.type import GraphQLObjectType
from pytest import raises

from graphene import Schema
from graphene.contrib.sqlalchemy.types import (SQLAlchemyNode,
                                               SQLAlchemyObjectType)
from graphene.core.fields import Field
from graphene.core.types.scalars import Int
from graphene.relay.fields import GlobalIDField
from tests.utils import assert_equal_lists

from .models import Article, Reporter

schema = Schema()


class Character(SQLAlchemyObjectType):
    '''Character description'''
    class Meta:
        model = Reporter


@schema.register
class Human(SQLAlchemyNode):
    '''Human description'''

    pub_date = Int()

    class Meta:
        model = Article
        exclude_fields = ('id', )


def test_sqlalchemy_interface():
    assert SQLAlchemyNode._meta.interface is True


# @patch('graphene.contrib.sqlalchemy.tests.models.Article.filter', return_value=Article(id=1))
# def test_sqlalchemy_get_node(get):
#     human = Human.get_node(1, None)
#     get.assert_called_with(id=1)
#     assert human.id == 1


def test_objecttype_registered():
    object_type = schema.T(Character)
    assert isinstance(object_type, GraphQLObjectType)
    assert Character._meta.model == Reporter
    assert_equal_lists(
        object_type.get_fields().keys(),
        ['articles', 'firstName', 'lastName', 'email', 'id']
    )


def test_sqlalchemynode_idfield():
    idfield = SQLAlchemyNode._meta.fields_map['id']
    assert isinstance(idfield, GlobalIDField)


def test_node_idfield():
    idfield = Human._meta.fields_map['id']
    assert isinstance(idfield, GlobalIDField)


def test_node_replacedfield():
    idfield = Human._meta.fields_map['pub_date']
    assert isinstance(idfield, Field)
    assert schema.T(idfield).type == schema.T(Int())


def test_interface_objecttype_init_none():
    h = Human()
    assert h._root is None


def test_interface_objecttype_init_good():
    instance = Article()
    h = Human(instance)
    assert h._root == instance


def test_interface_objecttype_init_unexpected():
    with raises(AssertionError) as excinfo:
        Human(object())
    assert str(excinfo.value) == "Human received a non-compatible instance (object) when expecting Article"


def test_object_type():
    object_type = schema.T(Human)
    Human._meta.fields_map
    assert Human._meta.interface is False
    assert isinstance(object_type, GraphQLObjectType)
    assert_equal_lists(
        object_type.get_fields().keys(),
        ['headline', 'id', 'reporter', 'reporterId', 'pubDate']
    )
    assert schema.T(SQLAlchemyNode) in object_type.get_interfaces()


def test_node_notinterface():
    assert Human._meta.interface is False
    assert SQLAlchemyNode in Human._meta.interfaces



from functools import wraps
from itertools import chain

from graphql.core.type import GraphQLArgument

from ...utils import ProxySnakeDict
from .base import ArgumentType, GroupNamedType, NamedType, OrderedType


class Argument(NamedType, OrderedType):

    def __init__(self, type, description=None, default=None,
                 name=None, _creation_counter=None):
        super(Argument, self).__init__(name=name, _creation_counter=_creation_counter)
        self.type = type
        self.description = description
        self.default = default

    def internal_type(self, schema):
        return GraphQLArgument(
            schema.T(self.type),
            self.default, self.description)

    def __repr__(self):
        return self.name


class ArgumentsGroup(GroupNamedType):

    def __init__(self, *args, **kwargs):
        arguments = to_arguments(*args, **kwargs)
        super(ArgumentsGroup, self).__init__(*arguments)


def to_arguments(*args, **kwargs):
    arguments = {}
    iter_arguments = chain(kwargs.items(), [(None, a) for a in args])

    for default_name, arg in iter_arguments:
        if isinstance(arg, Argument):
            argument = arg
        elif isinstance(arg, ArgumentType):
            argument = arg.as_argument()
        else:
            raise ValueError('Unknown argument %s=%r' % (default_name, arg))

        if default_name:
            argument.default_name = default_name

        name = argument.name or argument.default_name
        assert name, 'Argument in field must have a name'
        assert name not in arguments, 'Found more than one Argument with same name {}'.format(name)
        arguments[name] = argument

    return sorted(arguments.values())


def snake_case_args(resolver):
    @wraps(resolver)
    def wrapped_resolver(instance, args, info):
        return resolver(instance, ProxySnakeDict(args), info)

    return wrapped_resolver



# coding: utf-8
from ..str_converters import to_camel_case, to_snake_case, to_const


def test_snake_case():
    assert to_snake_case('snakesOnAPlane') == 'snakes_on_a_plane'
    assert to_snake_case('SnakesOnAPlane') == 'snakes_on_a_plane'
    assert to_snake_case('SnakesOnA_Plane') == 'snakes_on_a__plane'
    assert to_snake_case('snakes_on_a_plane') == 'snakes_on_a_plane'
    assert to_snake_case('snakes_on_a__plane') == 'snakes_on_a__plane'
    assert to_snake_case('IPhoneHysteria') == 'i_phone_hysteria'
    assert to_snake_case('iPhoneHysteria') == 'i_phone_hysteria'


def test_camel_case():
    assert to_camel_case('snakes_on_a_plane') == 'snakesOnAPlane'
    assert to_camel_case('snakes_on_a__plane') == 'snakesOnA_Plane'
    assert to_camel_case('i_phone_hysteria') == 'iPhoneHysteria'


def test_to_const():
    assert to_const('snakes $1. on a "#plane') == 'SNAKES_ON_A_PLANE'



import graphene
from graphene import relay
from graphene.contrib.sqlalchemy import (SQLAlchemyConnectionField,
                                         SQLAlchemyNode)
from models import Department as DepartmentModel
from models import Employee as EmployeeModel
from models import Role as RoleModel

schema = graphene.Schema()


@schema.register
class Department(SQLAlchemyNode):

    class Meta:
        model = DepartmentModel


@schema.register
class Employee(SQLAlchemyNode):

    class Meta:
        model = EmployeeModel


@schema.register
class Role(SQLAlchemyNode):

    class Meta:
        model = RoleModel
        identifier = 'role_id'


class Query(graphene.ObjectType):
    node = relay.NodeField(Employee)
    all_employees = SQLAlchemyConnectionField(Employee)
    all_roles = SQLAlchemyConnectionField(Role)
    role = relay.NodeField(Role)

schema.query = Query



from graphql.core.type import (GraphQLBoolean, GraphQLFloat, GraphQLID,
                               GraphQLInt, GraphQLString)

from graphene.core.schema import Schema

from ..scalars import ID, Boolean, Float, Int, String

schema = Schema()


def test_string_scalar():
    assert schema.T(String()) == GraphQLString


def test_int_scalar():
    assert schema.T(Int()) == GraphQLInt


def test_boolean_scalar():
    assert schema.T(Boolean()) == GraphQLBoolean


def test_id_scalar():
    assert schema.T(ID()) == GraphQLID


def test_float_scalar():
    assert schema.T(Float()) == GraphQLFloat



import pytest
from django.db import models
from py.test import raises

import graphene
from graphene.core.types.custom_scalars import DateTime, JSONString

from ..compat import (ArrayField, HStoreField, JSONField, MissingType,
                      RangeField)
from ..converter import convert_django_field, convert_django_field_with_choices
from ..fields import ConnectionOrListField, DjangoModelField
from .models import Article, Reporter


def assert_conversion(django_field, graphene_field, *args, **kwargs):
    field = django_field(help_text='Custom Help Text', *args, **kwargs)
    graphene_type = convert_django_field(field)
    assert isinstance(graphene_type, graphene_field)
    field = graphene_type.as_field()
    assert field.description == 'Custom Help Text'
    return field


def test_should_unknown_django_field_raise_exception():
    with raises(Exception) as excinfo:
        convert_django_field(None)
    assert 'Don\'t know how to convert the Django field' in str(excinfo.value)


def test_should_date_convert_string():
    assert_conversion(models.DateField, DateTime)


def test_should_char_convert_string():
    assert_conversion(models.CharField, graphene.String)


def test_should_text_convert_string():
    assert_conversion(models.TextField, graphene.String)


def test_should_email_convert_string():
    assert_conversion(models.EmailField, graphene.String)


def test_should_slug_convert_string():
    assert_conversion(models.SlugField, graphene.String)


def test_should_url_convert_string():
    assert_conversion(models.URLField, graphene.String)


def test_should_ipaddress_convert_string():
    assert_conversion(models.GenericIPAddressField, graphene.String)


def test_should_file_convert_string():
    assert_conversion(models.FileField, graphene.String)


def test_should_image_convert_string():
    assert_conversion(models.ImageField, graphene.String)


def test_should_auto_convert_id():
    assert_conversion(models.AutoField, graphene.ID, primary_key=True)


def test_should_positive_integer_convert_int():
    assert_conversion(models.PositiveIntegerField, graphene.Int)


def test_should_positive_small_convert_int():
    assert_conversion(models.PositiveSmallIntegerField, graphene.Int)


def test_should_small_integer_convert_int():
    assert_conversion(models.SmallIntegerField, graphene.Int)


def test_should_big_integer_convert_int():
    assert_conversion(models.BigIntegerField, graphene.Int)


def test_should_integer_convert_int():
    assert_conversion(models.IntegerField, graphene.Int)


def test_should_boolean_convert_boolean():
    field = assert_conversion(models.BooleanField, graphene.Boolean)
    assert field.required is True


def test_should_nullboolean_convert_boolean():
    field = assert_conversion(models.NullBooleanField, graphene.Boolean)
    assert field.required is False


def test_field_with_choices_convert_enum():
    field = models.CharField(help_text='Language', choices=(
        ('es', 'Spanish'),
        ('en', 'English')
    ))

    class TranslatedModel(models.Model):
        language = field

        class Meta:
            app_label = 'test'

    graphene_type = convert_django_field_with_choices(field)
    assert issubclass(graphene_type, graphene.Enum)
    assert graphene_type._meta.type_name == 'TEST_TRANSLATEDMODEL_LANGUAGE'
    assert graphene_type._meta.description == 'Language'
    assert graphene_type.__enum__.__members__['SPANISH'].value == 'es'
    assert graphene_type.__enum__.__members__['ENGLISH'].value == 'en'


def test_should_float_convert_float():
    assert_conversion(models.FloatField, graphene.Float)


def test_should_manytomany_convert_connectionorlist():
    graphene_type = convert_django_field(Reporter._meta.local_many_to_many[0])
    assert isinstance(graphene_type, ConnectionOrListField)
    assert isinstance(graphene_type.type, DjangoModelField)
    assert graphene_type.type.model == Reporter


def test_should_manytoone_convert_connectionorlist():
    # Django 1.9 uses 'rel', <1.9 uses 'related
    related = getattr(Reporter.articles, 'rel', None) or \
        getattr(Reporter.articles, 'related')
    graphene_type = convert_django_field(related)
    assert isinstance(graphene_type, ConnectionOrListField)
    assert isinstance(graphene_type.type, DjangoModelField)
    assert graphene_type.type.model == Article


def test_should_onetoone_convert_model():
    field = assert_conversion(models.OneToOneField, DjangoModelField, Article)
    assert field.type.model == Article


def test_should_foreignkey_convert_model():
    field = assert_conversion(models.ForeignKey, DjangoModelField, Article)
    assert field.type.model == Article


@pytest.mark.skipif(ArrayField is MissingType,
                    reason="ArrayField should exist")
def test_should_postgres_array_convert_list():
    field = assert_conversion(ArrayField, graphene.List, models.CharField(max_length=100))
    assert isinstance(field.type, graphene.List)
    assert isinstance(field.type.of_type, graphene.String)


@pytest.mark.skipif(ArrayField is MissingType,
                    reason="ArrayField should exist")
def test_should_postgres_array_multiple_convert_list():
    field = assert_conversion(ArrayField, graphene.List, ArrayField(models.CharField(max_length=100)))
    assert isinstance(field.type, graphene.List)
    assert isinstance(field.type.of_type, graphene.List)
    assert isinstance(field.type.of_type.of_type, graphene.String)


@pytest.mark.skipif(HStoreField is MissingType,
                    reason="HStoreField should exist")
def test_should_postgres_hstore_convert_string():
    assert_conversion(HStoreField, JSONString)


@pytest.mark.skipif(JSONField is MissingType,
                    reason="JSONField should exist")
def test_should_postgres_json_convert_string():
    assert_conversion(JSONField, JSONString)


@pytest.mark.skipif(RangeField is MissingType,
                    reason="RangeField should exist")
def test_should_postgres_range_convert_list():
    from django.contrib.postgres.fields import IntegerRangeField
    field = assert_conversion(IntegerRangeField, graphene.List)
    assert isinstance(field.type.of_type, graphene.Int)



import inspect

import six
from sqlalchemy.inspection import inspect as sqlalchemyinspect
from sqlalchemy.orm.exc import NoResultFound

from ...core.classtypes.objecttype import ObjectType, ObjectTypeMeta
from ...relay.types import Connection, Node, NodeMeta
from .converter import (convert_sqlalchemy_column,
                        convert_sqlalchemy_relationship)
from .options import SQLAlchemyOptions
from .utils import get_query, is_mapped


class SQLAlchemyObjectTypeMeta(ObjectTypeMeta):
    options_class = SQLAlchemyOptions

    def construct_fields(cls):
        only_fields = cls._meta.only_fields
        exclude_fields = cls._meta.exclude_fields
        already_created_fields = {f.attname for f in cls._meta.local_fields}
        inspected_model = sqlalchemyinspect(cls._meta.model)

        # Get all the columns for the relationships on the model
        for relationship in inspected_model.relationships:
            is_not_in_only = only_fields and relationship.key not in only_fields
            is_already_created = relationship.key in already_created_fields
            is_excluded = relationship.key in exclude_fields or is_already_created
            if is_not_in_only or is_excluded:
                # We skip this field if we specify only_fields and is not
                # in there. Or when we excldue this field in exclude_fields
                continue
            converted_relationship = convert_sqlalchemy_relationship(relationship)
            cls.add_to_class(relationship.key, converted_relationship)

        for column in inspected_model.columns:
            is_not_in_only = only_fields and column.name not in only_fields
            is_already_created = column.name in already_created_fields
            is_excluded = column.name in exclude_fields or is_already_created
            if is_not_in_only or is_excluded:
                # We skip this field if we specify only_fields and is not
                # in there. Or when we excldue this field in exclude_fields
                continue
            converted_column = convert_sqlalchemy_column(column)
            cls.add_to_class(column.name, converted_column)

    def construct(cls, *args, **kwargs):
        cls = super(SQLAlchemyObjectTypeMeta, cls).construct(*args, **kwargs)
        if not cls._meta.abstract:
            if not cls._meta.model:
                raise Exception(
                    'SQLAlchemy ObjectType %s must have a model in the Meta class attr' %
                    cls)
            elif not inspect.isclass(cls._meta.model) or not is_mapped(cls._meta.model):
                raise Exception('Provided model in %s is not a SQLAlchemy model' % cls)

            cls.construct_fields()
        return cls


class InstanceObjectType(ObjectType):

    class Meta:
        abstract = True

    def __init__(self, _root=None):
        super(InstanceObjectType, self).__init__(_root=_root)
        assert not self._root or isinstance(self._root, self._meta.model), (
            '{} received a non-compatible instance ({}) '
            'when expecting {}'.format(
                self.__class__.__name__,
                self._root.__class__.__name__,
                self._meta.model.__name__
            ))

    @property
    def instance(self):
        return self._root

    @instance.setter
    def instance(self, value):
        self._root = value


class SQLAlchemyObjectType(six.with_metaclass(
        SQLAlchemyObjectTypeMeta, InstanceObjectType)):

    class Meta:
        abstract = True


class SQLAlchemyConnection(Connection):
    pass


class SQLAlchemyNodeMeta(SQLAlchemyObjectTypeMeta, NodeMeta):
    pass


class NodeInstance(Node, InstanceObjectType):

    class Meta:
        abstract = True


class SQLAlchemyNode(six.with_metaclass(
        SQLAlchemyNodeMeta, NodeInstance)):

    class Meta:
        abstract = True

    def to_global_id(self):
        id_ = getattr(self.instance, self._meta.identifier)
        return self.global_id(id_)

    @classmethod
    def get_node(cls, id, info=None):
        try:
            model = cls._meta.model
            identifier = cls._meta.identifier
            query = get_query(model, info)
            instance = query.filter(getattr(model, identifier) == id).one()
            return cls(instance)
        except NoResultFound:
            return None



# flake8: noqa
"""Python Enumerations"""
from __future__ import absolute_import

__all__ = ['Enum', 'IntEnum', 'unique']

try:
    from enum import Enum, IntEnum, unique
except ImportError:
    import sys as _sys

    version = 1, 1, 2

    pyver = float('%s.%s' % _sys.version_info[:2])

    try:
        any
    except NameError:
        def any(iterable):
            for element in iterable:
                if element:
                    return True
            return False

    try:
        from collections import OrderedDict
    except ImportError:
        OrderedDict = None

    try:
        basestring
    except NameError:
        # In Python 2 basestring is the ancestor of both str and unicode
        # in Python 3 it's just str, but was missing in 3.1
        basestring = str

    try:
        unicode
    except NameError:
        # In Python 3 unicode no longer exists (it's just str)
        unicode = str

    class _RouteClassAttributeToGetattr(object):
        """Route attribute access on a class to __getattr__.

        This is a descriptor, used to define attributes that act differently when
        accessed through an instance and through a class.  Instance access remains
        normal, but access to an attribute through a class will be routed to the
        class's __getattr__ method; this is done by raising AttributeError.

        """

        def __init__(self, fget=None):
            self.fget = fget

        def __get__(self, instance, ownerclass=None):
            if instance is None:
                raise AttributeError()
            return self.fget(instance)

        def __set__(self, instance, value):
            raise AttributeError("can't set attribute")

        def __delete__(self, instance):
            raise AttributeError("can't delete attribute")

    def _is_descriptor(obj):
        """Returns True if obj is a descriptor, False otherwise."""
        return (
            hasattr(obj, '__get__') or
            hasattr(obj, '__set__') or
            hasattr(obj, '__delete__'))

    def _is_dunder(name):
        """Returns True if a __dunder__ name, False otherwise."""
        return (name[:2] == name[-2:] == '__' and
                name[2:3] != '_' and
                name[-3:-2] != '_' and
                len(name) > 4)

    def _is_sunder(name):
        """Returns True if a _sunder_ name, False otherwise."""
        return (name[0] == name[-1] == '_' and
                name[1:2] != '_' and
                name[-2:-1] != '_' and
                len(name) > 2)

    def _make_class_unpicklable(cls):
        """Make the given class un-picklable."""

        def _break_on_call_reduce(self, protocol=None):
            raise TypeError('%r cannot be pickled' % self)
        cls.__reduce_ex__ = _break_on_call_reduce
        cls.__module__ = '<unknown>'

    class _EnumDict(dict):
        """Track enum member order and ensure member names are not reused.

        EnumMeta will use the names found in self._member_names as the
        enumeration member names.

        """

        def __init__(self):
            super(_EnumDict, self).__init__()
            self._member_names = []

        def __setitem__(self, key, value):
            """Changes anything not dundered or not a descriptor.

            If a descriptor is added with the same name as an enum member, the name
            is removed from _member_names (this may leave a hole in the numerical
            sequence of values).

            If an enum member name is used twice, an error is raised; duplicate
            values are not checked for.

            Single underscore (sunder) names are reserved.

            Note:   in 3.x __order__ is simply discarded as a not necessary piece
                    leftover from 2.x

            """
            if pyver >= 3.0 and key == '__order__':
                return
            if _is_sunder(key):
                raise ValueError('_names_ are reserved for future Enum use')
            elif _is_dunder(key):
                pass
            elif key in self._member_names:
                # descriptor overwriting an enum?
                raise TypeError('Attempted to reuse key: %r' % key)
            elif not _is_descriptor(value):
                if key in self:
                    # enum overwriting a descriptor?
                    raise TypeError('Key already defined as: %r' % self[key])
                self._member_names.append(key)
            super(_EnumDict, self).__setitem__(key, value)

    # Dummy value for Enum as EnumMeta explicity checks for it, but of course until
    # EnumMeta finishes running the first time the Enum class doesn't exist.  This
    # is also why there are checks in EnumMeta like `if Enum is not None`
    Enum = None

    class EnumMeta(type):
        """Metaclass for Enum"""
        @classmethod
        def __prepare__(metacls, cls, bases):
            return _EnumDict()

        def __new__(metacls, cls, bases, classdict):
            # an Enum class is final once enumeration items have been defined; it
            # cannot be mixed with other types (int, float, etc.) if it has an
            # inherited __new__ unless a new __new__ is defined (or the resulting
            # class will fail).
            if isinstance(classdict, dict):
                original_dict = classdict
                classdict = _EnumDict()
                for k, v in original_dict.items():
                    classdict[k] = v

            member_type, first_enum = metacls._get_mixins_(bases)
            __new__, save_new, use_args = metacls._find_new_(classdict, member_type,
                                                             first_enum)
            # save enum items into separate mapping so they don't get baked into
            # the new class
            members = dict((k, classdict[k]) for k in classdict._member_names)
            for name in classdict._member_names:
                del classdict[name]

            # py2 support for definition order
            __order__ = classdict.get('__order__')
            if __order__ is None:
                if pyver < 3.0:
                    try:
                        __order__ = [name for (name, value) in sorted(members.items(), key=lambda item: item[1])]
                    except TypeError:
                        __order__ = [name for name in sorted(members.keys())]
                else:
                    __order__ = classdict._member_names
            else:
                del classdict['__order__']
                if pyver < 3.0:
                    __order__ = __order__.replace(',', ' ').split()
                    aliases = [name for name in members if name not in __order__]
                    __order__ += aliases

            # check for illegal enum names (any others?)
            invalid_names = set(members) & set(['mro'])
            if invalid_names:
                raise ValueError('Invalid enum member name(s): %s' % (
                    ', '.join(invalid_names), ))

            # save attributes from super classes so we know if we can take
            # the shortcut of storing members in the class dict
            base_attributes = set([a for b in bases for a in b.__dict__])
            # create our new Enum type
            enum_class = super(EnumMeta, metacls).__new__(metacls, cls, bases, classdict)
            enum_class._member_names_ = []               # names in random order
            if OrderedDict is not None:
                enum_class._member_map_ = OrderedDict()
            else:
                enum_class._member_map_ = {}             # name->value map
            enum_class._member_type_ = member_type

            # Reverse value->name map for hashable values.
            enum_class._value2member_map_ = {}

            # instantiate them, checking for duplicates as we go
            # we instantiate first instead of checking for duplicates first in case
            # a custom __new__ is doing something funky with the values -- such as
            # auto-numbering ;)
            if __new__ is None:
                __new__ = enum_class.__new__
            for member_name in __order__:
                value = members[member_name]
                if not isinstance(value, tuple):
                    args = (value, )
                else:
                    args = value
                if member_type is tuple:   # special case for tuple enums
                    args = (args, )     # wrap it one more time
                if not use_args or not args:
                    enum_member = __new__(enum_class)
                    if not hasattr(enum_member, '_value_'):
                        enum_member._value_ = value
                else:
                    enum_member = __new__(enum_class, *args)
                    if not hasattr(enum_member, '_value_'):
                        enum_member._value_ = member_type(*args)
                value = enum_member._value_
                enum_member._name_ = member_name
                enum_member.__objclass__ = enum_class
                enum_member.__init__(*args)
                # If another member with the same value was already defined, the
                # new member becomes an alias to the existing one.
                for name, canonical_member in enum_class._member_map_.items():
                    if canonical_member.value == enum_member._value_:
                        enum_member = canonical_member
                        break
                else:
                    # Aliases don't appear in member names (only in __members__).
                    enum_class._member_names_.append(member_name)
                # performance boost for any member that would not shadow
                # a DynamicClassAttribute (aka _RouteClassAttributeToGetattr)
                if member_name not in base_attributes:
                    setattr(enum_class, member_name, enum_member)
                # now add to _member_map_
                enum_class._member_map_[member_name] = enum_member
                try:
                    # This may fail if value is not hashable. We can't add the value
                    # to the map, and by-value lookups for this value will be
                    # linear.
                    enum_class._value2member_map_[value] = enum_member
                except TypeError:
                    pass

            # If a custom type is mixed into the Enum, and it does not know how
            # to pickle itself, pickle.dumps will succeed but pickle.loads will
            # fail.  Rather than have the error show up later and possibly far
            # from the source, sabotage the pickle protocol for this class so
            # that pickle.dumps also fails.
            #
            # However, if the new class implements its own __reduce_ex__, do not
            # sabotage -- it's on them to make sure it works correctly.  We use
            # __reduce_ex__ instead of any of the others as it is preferred by
            # pickle over __reduce__, and it handles all pickle protocols.
            unpicklable = False
            if '__reduce_ex__' not in classdict:
                if member_type is not object:
                    methods = ('__getnewargs_ex__', '__getnewargs__',
                               '__reduce_ex__', '__reduce__')
                    if not any(m in member_type.__dict__ for m in methods):
                        _make_class_unpicklable(enum_class)
                        unpicklable = True

            # double check that repr and friends are not the mixin's or various
            # things break (such as pickle)
            for name in ('__repr__', '__str__', '__format__', '__reduce_ex__'):
                class_method = getattr(enum_class, name)
                getattr(member_type, name, None)
                enum_method = getattr(first_enum, name, None)
                if name not in classdict and class_method is not enum_method:
                    if name == '__reduce_ex__' and unpicklable:
                        continue
                    setattr(enum_class, name, enum_method)

            # method resolution and int's are not playing nice
            # Python's less than 2.6 use __cmp__

            if pyver < 2.6:

                if issubclass(enum_class, int):
                    setattr(enum_class, '__cmp__', getattr(int, '__cmp__'))

            elif pyver < 3.0:

                if issubclass(enum_class, int):
                    for method in (
                            '__le__',
                            '__lt__',
                            '__gt__',
                            '__ge__',
                            '__eq__',
                            '__ne__',
                            '__hash__',
                    ):
                        setattr(enum_class, method, getattr(int, method))

            # replace any other __new__ with our own (as long as Enum is not None,
            # anyway) -- again, this is to support pickle
            if Enum is not None:
                # if the user defined their own __new__, save it before it gets
                # clobbered in case they subclass later
                if save_new:
                    setattr(enum_class, '__member_new__', enum_class.__dict__['__new__'])
                setattr(enum_class, '__new__', Enum.__dict__['__new__'])
            return enum_class

        def __call__(cls, value, names=None, module=None, type=None, start=1):
            """Either returns an existing member, or creates a new enum class.

            This method is used both when an enum class is given a value to match
            to an enumeration member (i.e. Color(3)) and for the functional API
            (i.e. Color = Enum('Color', names='red green blue')).

            When used for the functional API: `module`, if set, will be stored in
            the new class' __module__ attribute; `type`, if set, will be mixed in
            as the first base class.

            Note: if `module` is not set this routine will attempt to discover the
            calling module by walking the frame stack; if this is unsuccessful
            the resulting class will not be pickleable.

            """
            if names is None:  # simple value lookup
                return cls.__new__(cls, value)
            # otherwise, functional API: we're creating a new Enum type
            return cls._create_(value, names, module=module, type=type, start=start)

        def __contains__(cls, member):
            return isinstance(member, cls) and member.name in cls._member_map_

        def __delattr__(cls, attr):
            # nicer error message when someone tries to delete an attribute
            # (see issue19025).
            if attr in cls._member_map_:
                raise AttributeError(
                    "%s: cannot delete Enum member." % cls.__name__)
            super(EnumMeta, cls).__delattr__(attr)

        def __dir__(self):
            return (['__class__', '__doc__', '__members__', '__module__'] +
                    self._member_names_)

        @property
        def __members__(cls):
            """Returns a mapping of member name->value.

            This mapping lists all enum members, including aliases. Note that this
            is a copy of the internal mapping.

            """
            return cls._member_map_.copy()

        def __getattr__(cls, name):
            """Return the enum member matching `name`

            We use __getattr__ instead of descriptors or inserting into the enum
            class' __dict__ in order to support `name` and `value` being both
            properties for enum members (which live in the class' __dict__) and
            enum members themselves.

            """
            if _is_dunder(name):
                raise AttributeError(name)
            try:
                return cls._member_map_[name]
            except KeyError:
                raise AttributeError(name)

        def __getitem__(cls, name):
            return cls._member_map_[name]

        def __iter__(cls):
            return (cls._member_map_[name] for name in cls._member_names_)

        def __reversed__(cls):
            return (cls._member_map_[name] for name in reversed(cls._member_names_))

        def __len__(cls):
            return len(cls._member_names_)

        def __repr__(cls):
            return "<enum %r>" % cls.__name__

        def __setattr__(cls, name, value):
            """Block attempts to reassign Enum members.

            A simple assignment to the class namespace only changes one of the
            several possible ways to get an Enum member from the Enum class,
            resulting in an inconsistent Enumeration.

            """
            member_map = cls.__dict__.get('_member_map_', {})
            if name in member_map:
                raise AttributeError('Cannot reassign members.')
            super(EnumMeta, cls).__setattr__(name, value)

        def _create_(cls, class_name, names=None, module=None, type=None, start=1):
            """Convenience method to create a new Enum class.

            `names` can be:

            * A string containing member names, separated either with spaces or
              commas.  Values are auto-numbered from 1.
            * An iterable of member names.  Values are auto-numbered from 1.
            * An iterable of (member name, value) pairs.
            * A mapping of member name -> value.

            """
            if pyver < 3.0:
                # if class_name is unicode, attempt a conversion to ASCII
                if isinstance(class_name, unicode):
                    try:
                        class_name = class_name.encode('ascii')
                    except UnicodeEncodeError:
                        raise TypeError('%r is not representable in ASCII' % class_name)
            metacls = cls.__class__
            if type is None:
                bases = (cls, )
            else:
                bases = (type, cls)
            classdict = metacls.__prepare__(class_name, bases)
            __order__ = []

            # special processing needed for names?
            if isinstance(names, basestring):
                names = names.replace(',', ' ').split()
            if isinstance(names, (tuple, list)) and isinstance(names[0], basestring):
                names = [(e, i + start) for (i, e) in enumerate(names)]

            # Here, names is either an iterable of (name, value) or a mapping.
            item = None  # in case names is empty
            for item in names:
                if isinstance(item, basestring):
                    member_name, member_value = item, names[item]
                else:
                    member_name, member_value = item
                classdict[member_name] = member_value
                __order__.append(member_name)
            # only set __order__ in classdict if name/value was not from a mapping
            if not isinstance(item, basestring):
                classdict['__order__'] = ' '.join(__order__)
            enum_class = metacls.__new__(metacls, class_name, bases, classdict)

            # TODO: replace the frame hack if a blessed way to know the calling
            # module is ever developed
            if module is None:
                try:
                    module = _sys._getframe(2).f_globals['__name__']
                except (AttributeError, ValueError):
                    pass
            if module is None:
                _make_class_unpicklable(enum_class)
            else:
                enum_class.__module__ = module

            return enum_class

        @staticmethod
        def _get_mixins_(bases):
            """Returns the type for creating enum members, and the first inherited
            enum class.

            bases: the tuple of bases that was given to __new__

            """
            if not bases or Enum is None:
                return object, Enum

            # double check that we are not subclassing a class with existing
            # enumeration members; while we're at it, see if any other data
            # type has been mixed in so we can use the correct __new__
            member_type = first_enum = None
            for base in bases:
                if (base is not Enum and
                        issubclass(base, Enum) and
                        base._member_names_):
                    raise TypeError("Cannot extend enumerations")
            # base is now the last base in bases
            if not issubclass(base, Enum):
                raise TypeError("new enumerations must be created as "
                                "`ClassName([mixin_type,] enum_type)`")

            # get correct mix-in type (either mix-in type of Enum subclass, or
            # first base if last base is Enum)
            if not issubclass(bases[0], Enum):
                member_type = bases[0]     # first data type
                first_enum = bases[-1]  # enum type
            else:
                for base in bases[0].__mro__:
                    # most common: (IntEnum, int, Enum, object)
                    # possible:    (<Enum 'AutoIntEnum'>, <Enum 'IntEnum'>,
                    #               <class 'int'>, <Enum 'Enum'>,
                    #               <class 'object'>)
                    if issubclass(base, Enum):
                        if first_enum is None:
                            first_enum = base
                    else:
                        if member_type is None:
                            member_type = base

            return member_type, first_enum

        if pyver < 3.0:
            @staticmethod
            def _find_new_(classdict, member_type, first_enum):
                """Returns the __new__ to be used for creating the enum members.

                classdict: the class dictionary given to __new__
                member_type: the data type whose __new__ will be used by default
                first_enum: enumeration to check for an overriding __new__

                """
                # now find the correct __new__, checking to see of one was defined
                # by the user; also check earlier enum classes in case a __new__ was
                # saved as __member_new__
                __new__ = classdict.get('__new__', None)
                if __new__:
                    return None, True, True      # __new__, save_new, use_args

                N__new__ = getattr(None, '__new__')
                O__new__ = getattr(object, '__new__')
                if Enum is None:
                    E__new__ = N__new__
                else:
                    E__new__ = Enum.__dict__['__new__']
                # check all possibles for __member_new__ before falling back to
                # __new__
                for method in ('__member_new__', '__new__'):
                    for possible in (member_type, first_enum):
                        try:
                            target = possible.__dict__[method]
                        except (AttributeError, KeyError):
                            target = getattr(possible, method, None)
                        if target not in [
                                None,
                                N__new__,
                                O__new__,
                                E__new__,
                        ]:
                            if method == '__member_new__':
                                classdict['__new__'] = target
                                return None, False, True
                            if isinstance(target, staticmethod):
                                target = target.__get__(member_type)
                            __new__ = target
                            break
                    if __new__ is not None:
                        break
                else:
                    __new__ = object.__new__

                # if a non-object.__new__ is used then whatever value/tuple was
                # assigned to the enum member name will be passed to __new__ and to the
                # new enum member's __init__
                if __new__ is object.__new__:
                    use_args = False
                else:
                    use_args = True

                return __new__, False, use_args
        else:
            @staticmethod
            def _find_new_(classdict, member_type, first_enum):
                """Returns the __new__ to be used for creating the enum members.

                classdict: the class dictionary given to __new__
                member_type: the data type whose __new__ will be used by default
                first_enum: enumeration to check for an overriding __new__

                """
                # now find the correct __new__, checking to see of one was defined
                # by the user; also check earlier enum classes in case a __new__ was
                # saved as __member_new__
                __new__ = classdict.get('__new__', None)

                # should __new__ be saved as __member_new__ later?
                save_new = __new__ is not None

                if __new__ is None:
                    # check all possibles for __member_new__ before falling back to
                    # __new__
                    for method in ('__member_new__', '__new__'):
                        for possible in (member_type, first_enum):
                            target = getattr(possible, method, None)
                            if target not in (
                                    None,
                                    None.__new__,
                                    object.__new__,
                                    Enum.__new__,
                            ):
                                __new__ = target
                                break
                        if __new__ is not None:
                            break
                    else:
                        __new__ = object.__new__

                # if a non-object.__new__ is used then whatever value/tuple was
                # assigned to the enum member name will be passed to __new__ and to the
                # new enum member's __init__
                if __new__ is object.__new__:
                    use_args = False
                else:
                    use_args = True

                return __new__, save_new, use_args

    ########################################################
    # In order to support Python 2 and 3 with a single
    # codebase we have to create the Enum methods separately
    # and then use the `type(name, bases, dict)` method to
    # create the class.
    ########################################################
    temp_enum_dict = {}
    temp_enum_dict['__doc__'] = "Generic enumeration.\n\n    Derive from this class to define new enumerations.\n\n"

    def __new__(cls, value):
        # all enum instances are actually created during class construction
        # without calling this method; this method is called by the metaclass'
        # __call__ (i.e. Color(3) ), and by pickle
        if isinstance(value, cls):
            # For lookups like Color(Color.red)
            value = value.value
            # return value
        # by-value search for a matching enum member
        # see if it's in the reverse mapping (for hashable values)
        try:
            if value in cls._value2member_map_:
                return cls._value2member_map_[value]
        except TypeError:
            # not there, now do long search -- O(n) behavior
            for member in cls._member_map_.values():
                if member.value == value:
                    return member
        raise ValueError("%s is not a valid %s" % (value, cls.__name__))
    temp_enum_dict['__new__'] = __new__
    del __new__

    def __repr__(self):
        return "<%s.%s: %r>" % (
            self.__class__.__name__, self._name_, self._value_)
    temp_enum_dict['__repr__'] = __repr__
    del __repr__

    def __str__(self):
        return "%s.%s" % (self.__class__.__name__, self._name_)
    temp_enum_dict['__str__'] = __str__
    del __str__

    if pyver >= 3.0:
        def __dir__(self):
            added_behavior = [
                m
                for cls in self.__class__.mro()
                for m in cls.__dict__
                if m[0] != '_' and m not in self._member_map_
            ]
            return (['__class__', '__doc__', '__module__', ] + added_behavior)
        temp_enum_dict['__dir__'] = __dir__
        del __dir__

    def __format__(self, format_spec):
        # mixed-in Enums should use the mixed-in type's __format__, otherwise
        # we can get strange results with the Enum name showing up instead of
        # the value

        # pure Enum branch
        if self._member_type_ is object:
            cls = str
            val = str(self)
        # mix-in branch
        else:
            cls = self._member_type_
            val = self.value
        return cls.__format__(val, format_spec)
    temp_enum_dict['__format__'] = __format__
    del __format__

    ####################################
    # Python's less than 2.6 use __cmp__

    if pyver < 2.6:

        def __cmp__(self, other):
            if isinstance(other, self.__class__):
                if self is other:
                    return 0
                return -1
            return NotImplemented
            raise TypeError("unorderable types: %s() and %s()" % (self.__class__.__name__, other.__class__.__name__))
        temp_enum_dict['__cmp__'] = __cmp__
        del __cmp__

    else:

        def __le__(self, other):
            raise TypeError("unorderable types: %s() <= %s()" % (self.__class__.__name__, other.__class__.__name__))
        temp_enum_dict['__le__'] = __le__
        del __le__

        def __lt__(self, other):
            raise TypeError("unorderable types: %s() < %s()" % (self.__class__.__name__, other.__class__.__name__))
        temp_enum_dict['__lt__'] = __lt__
        del __lt__

        def __ge__(self, other):
            raise TypeError("unorderable types: %s() >= %s()" % (self.__class__.__name__, other.__class__.__name__))
        temp_enum_dict['__ge__'] = __ge__
        del __ge__

        def __gt__(self, other):
            raise TypeError("unorderable types: %s() > %s()" % (self.__class__.__name__, other.__class__.__name__))
        temp_enum_dict['__gt__'] = __gt__
        del __gt__

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self is other
        return NotImplemented
    temp_enum_dict['__eq__'] = __eq__
    del __eq__

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return self is not other
        return NotImplemented
    temp_enum_dict['__ne__'] = __ne__
    del __ne__

    def __hash__(self):
        return hash(self._name_)
    temp_enum_dict['__hash__'] = __hash__
    del __hash__

    # TODO: enable once Python 3.6 is released
    # def __bool__(self):
    #     return bool(self._value_)
    # if pyver < 3.0:
    #     temp_enum_dict['__nonzero__'] = __bool__
    # else:
    #     temp_enum_dict['__bool__'] = __bool__
    #     del __bool__

    def __reduce_ex__(self, proto):
        return self.__class__, (self._value_, )
    temp_enum_dict['__reduce_ex__'] = __reduce_ex__
    del __reduce_ex__

    # _RouteClassAttributeToGetattr is used to provide access to the `name`
    # and `value` properties of enum members while keeping some measure of
    # protection from modification, while still allowing for an enumeration
    # to have members named `name` and `value`.  This works because enumeration
    # members are not set directly on the enum class -- __getattr__ is
    # used to look them up.

    @_RouteClassAttributeToGetattr
    def name(self):
        return self._name_
    temp_enum_dict['name'] = name
    del name

    @_RouteClassAttributeToGetattr
    def value(self):
        return self._value_
    temp_enum_dict['value'] = value
    del value

    @classmethod
    def _convert(cls, name, module, filter, source=None):
        """
        Create a new Enum subclass that replaces a collection of global constants
        """
        # convert all constants from source (or module) that pass filter() to
        # a new Enum called name, and export the enum and its members back to
        # module;
        # also, replace the __reduce_ex__ method so unpickling works in
        # previous Python versions
        module_globals = vars(_sys.modules[module])
        if source:
            source = vars(source)
        else:
            source = module_globals
        members = dict((name, value) for name, value in source.items() if filter(name))
        cls = cls(name, members, module=module)
        cls.__reduce_ex__ = _reduce_ex_by_name
        module_globals.update(cls.__members__)
        module_globals[name] = cls
        return cls
    temp_enum_dict['_convert'] = _convert
    del _convert

    Enum = EnumMeta('Enum', (object, ), temp_enum_dict)
    del temp_enum_dict

    # Enum has now been created
    ###########################

    class IntEnum(int, Enum):
        """Enum where members are also (and must be) ints"""

    def _reduce_ex_by_name(self, proto):
        return self.name

    def unique(enumeration):
        """Class decorator that ensures only unique members exist in an enumeration."""
        duplicates = []
        for name, member in enumeration.__members__.items():
            if name != member.name:
                duplicates.append((name, member.name))
        if duplicates:
            duplicate_names = ', '.join(
                ["%s -> %s" % (alias, name) for (alias, name) in duplicates]
            )
            raise ValueError('duplicate names found in %r: %s' %
                             (enumeration, duplicate_names)
                             )
        return enumeration



from graphql.core.type import GraphQLEnumType

from graphene.core.schema import Schema

from ..enum import Enum
from ..objecttype import ObjectType


def test_enum():
    class RGB(Enum):
        '''RGB enum description'''
        RED = 0
        GREEN = 1
        BLUE = 2

    schema = Schema()

    object_type = schema.T(RGB)
    assert isinstance(object_type, GraphQLEnumType)
    assert RGB._meta.type_name == 'RGB'
    assert RGB._meta.description == 'RGB enum description'
    assert RGB.RED == 0
    assert RGB.GREEN == 1
    assert RGB.BLUE == 2


def test_enum_values():
    RGB = Enum('RGB', dict(RED=0, GREEN=1, BLUE=2), description='RGB enum description')

    schema = Schema()

    object_type = schema.T(RGB)
    assert isinstance(object_type, GraphQLEnumType)
    assert RGB._meta.type_name == 'RGB'
    assert RGB._meta.description == 'RGB enum description'
    assert RGB.RED == 0
    assert RGB.GREEN == 1
    assert RGB.BLUE == 2


def test_enum_instance():
    RGB = Enum('RGB', dict(RED=0, GREEN=1, BLUE=2))
    RGB_field = RGB(description='RGB enum description')

    class ObjectWithColor(ObjectType):
        color = RGB_field

    object_field = ObjectWithColor._meta.fields_map['color']
    assert object_field.description == 'RGB enum description'



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

import graphene
from graphene import relay
from graphene.contrib.sqlalchemy import (SQLAlchemyConnectionField,
                                         SQLAlchemyNode, SQLAlchemyObjectType)

from .models import Article, Base, Reporter, Editor

db = create_engine('sqlite:///test_sqlalchemy.sqlite3')


@pytest.yield_fixture(scope='function')
def session():
    connection = db.engine.connect()
    transaction = connection.begin()
    Base.metadata.create_all(connection)

    # options = dict(bind=connection, binds={})
    session_factory = sessionmaker(bind=connection)
    session = scoped_session(session_factory)

    yield session

    # Finalize test here
    transaction.rollback()
    connection.close()
    session.remove()


def setup_fixtures(session):
    reporter = Reporter(first_name='ABA', last_name='X')
    session.add(reporter)
    reporter2 = Reporter(first_name='ABO', last_name='Y')
    session.add(reporter2)
    article = Article(headline='Hi!')
    session.add(article)
    editor = Editor(name="John")
    session.add(editor)
    session.commit()


def test_should_query_well(session):
    setup_fixtures(session)

    class ReporterType(SQLAlchemyObjectType):

        class Meta:
            model = Reporter

    class Query(graphene.ObjectType):
        reporter = graphene.Field(ReporterType)
        reporters = ReporterType.List()

        def resolve_reporter(self, *args, **kwargs):
            return session.query(Reporter).first()

        def resolve_reporters(self, *args, **kwargs):
            return session.query(Reporter)

    query = '''
        query ReporterQuery {
          reporter {
            firstName,
            lastName,
            email
          }
          reporters {
            firstName
          }
        }
    '''
    expected = {
        'reporter': {
            'firstName': 'ABA',
            'lastName': 'X',
            'email': None
        },
        'reporters': [{
            'firstName': 'ABA',
        }, {
            'firstName': 'ABO',
        }]
    }
    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_should_node(session):
    setup_fixtures(session)

    class ReporterNode(SQLAlchemyNode):

        class Meta:
            model = Reporter

        @classmethod
        def get_node(cls, id, info):
            return Reporter(id=2, first_name='Cookie Monster')

        def resolve_articles(self, *args, **kwargs):
            return [Article(headline='Hi!')]

    class ArticleNode(SQLAlchemyNode):

        class Meta:
            model = Article

        # @classmethod
        # def get_node(cls, id, info):
        #     return Article(id=1, headline='Article node')

    class Query(graphene.ObjectType):
        node = relay.NodeField()
        reporter = graphene.Field(ReporterNode)
        article = graphene.Field(ArticleNode)
        all_articles = SQLAlchemyConnectionField(ArticleNode)

        def resolve_reporter(self, *args, **kwargs):
            return Reporter(id=1, first_name='ABA', last_name='X')

        def resolve_article(self, *args, **kwargs):
            return Article(id=1, headline='Article node')

    query = '''
        query ReporterQuery {
          reporter {
            id,
            firstName,
            articles {
              edges {
                node {
                  headline
                }
              }
            }
            lastName,
            email
          }
          allArticles {
            edges {
              node {
                headline
              }
            }
          }
          myArticle: node(id:"QXJ0aWNsZU5vZGU6MQ==") {
            id
            ... on ReporterNode {
                firstName
            }
            ... on ArticleNode {
                headline
            }
          }
        }
    '''
    expected = {
        'reporter': {
            'id': 'UmVwb3J0ZXJOb2RlOjE=',
            'firstName': 'ABA',
            'lastName': 'X',
            'email': None,
            'articles': {
                'edges': [{
                  'node': {
                      'headline': 'Hi!'
                  }
                }]
            },
        },
        'allArticles': {
            'edges': [{
                'node': {
                    'headline': 'Hi!'
                }
            }]
        },
        'myArticle': {
            'id': 'QXJ0aWNsZU5vZGU6MQ==',
            'headline': 'Hi!'
        }
    }
    schema = graphene.Schema(query=Query, session=session)
    result = schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_should_custom_identifier(session):
    setup_fixtures(session)

    class EditorNode(SQLAlchemyNode):

        class Meta:
            model = Editor
            identifier = "editor_id"

    class Query(graphene.ObjectType):
        node = relay.NodeField(EditorNode)
        all_editors = SQLAlchemyConnectionField(EditorNode)

    query = '''
        query EditorQuery {
          allEditors {
            edges {
                node {
                    id,
                    name
                }
            }
          },
          node(id: "RWRpdG9yTm9kZTox") {
            name
          }
        }
    '''
    expected = {
        'allEditors': {
            'edges': [{
                'node': {
                    'id': 'RWRpdG9yTm9kZTox',
                    'name': 'John'
                }
            }]
        },
        'node': {
            'name': 'John'
        }
    }

    schema = graphene.Schema(query=Query, session=session)
    result = schema.execute(query)
    assert not result.errors
    assert result.data == expected



from collections import OrderedDict
from functools import partial, total_ordering

import six


class InstanceType(object):

    def internal_type(self, schema):
        raise NotImplementedError("internal_type for type {} is not implemented".format(self.__class__.__name__))


class MountType(InstanceType):
    parent = None

    def mount(self, cls):
        self.parent = cls


class LazyType(MountType):

    def __init__(self, type):
        self.type = type

    @property
    def is_self(self):
        return self.type == 'self'

    def internal_type(self, schema):
        type = None
        if callable(self.type):
            type = self.type(self.parent)
        elif isinstance(self.type, six.string_types):
            if self.is_self:
                type = self.parent
            else:
                type = schema.get_type(self.type)
        assert type, 'Type in %s %r cannot be none' % (self.type, self.parent)
        return schema.T(type)


@total_ordering
class OrderedType(MountType):
    creation_counter = 0

    def __init__(self, _creation_counter=None):
        self.creation_counter = _creation_counter or self.gen_counter()

    @staticmethod
    def gen_counter():
        counter = OrderedType.creation_counter
        OrderedType.creation_counter += 1
        return counter

    def __eq__(self, other):
        # Needed for @total_ordering
        if isinstance(self, type(other)):
            return self.creation_counter == other.creation_counter
        return NotImplemented

    def __lt__(self, other):
        # This is needed because bisect does not take a comparison function.
        if isinstance(other, OrderedType):
            return self.creation_counter < other.creation_counter
        return NotImplemented

    def __gt__(self, other):
        # This is needed because bisect does not take a comparison function.
        if isinstance(other, OrderedType):
            return self.creation_counter > other.creation_counter
        return NotImplemented

    def __hash__(self):
        return hash((self.creation_counter))


class MirroredType(OrderedType):

    def __init__(self, *args, **kwargs):
        _creation_counter = kwargs.pop('_creation_counter', None)
        super(MirroredType, self).__init__(_creation_counter=_creation_counter)
        self.args = args
        self.kwargs = kwargs

    @property
    def List(self):  # noqa
        from .definitions import List
        return List(self, *self.args, **self.kwargs)

    @property
    def NonNull(self):  # noqa
        from .definitions import NonNull
        return NonNull(self, *self.args, **self.kwargs)


class ArgumentType(MirroredType):

    def as_argument(self):
        from .argument import Argument
        return Argument(
            self, _creation_counter=self.creation_counter, *self.args, **self.kwargs)


class FieldType(MirroredType):

    def contribute_to_class(self, cls, name):
        from ..classtypes.base import FieldsClassType
        from ..classtypes.inputobjecttype import InputObjectType
        if issubclass(cls, (InputObjectType)):
            inputfield = self.as_inputfield()
            return inputfield.contribute_to_class(cls, name)
        elif issubclass(cls, (FieldsClassType)):
            field = self.as_field()
            return field.contribute_to_class(cls, name)

    def as_field(self):
        from .field import Field
        return Field(self, _creation_counter=self.creation_counter,
                     *self.args, **self.kwargs)

    def as_inputfield(self):
        from .field import InputField
        return InputField(
            self, _creation_counter=self.creation_counter, *self.args, **self.kwargs)


class MountedType(FieldType, ArgumentType):
    pass


class NamedType(InstanceType):

    def __init__(self, name=None, default_name=None, *args, **kwargs):
        self.name = name
        self.default_name = None
        super(NamedType, self).__init__(*args, **kwargs)


class GroupNamedType(InstanceType):

    def __init__(self, *types):
        self.types = types

    def get_named_type(self, schema, type):
        name = type.name or schema.get_default_namedtype_name(type.default_name)
        return name, schema.T(type)

    def iter_types(self, schema):
        return map(partial(self.get_named_type, schema), self.types)

    def internal_type(self, schema):
        return OrderedDict(self.iter_types(schema))

    def __len__(self):
        return len(self.types)

    def __iter__(self):
        return iter(self.types)

    def __contains__(self, *args):
        return self.types.__contains__(*args)

    def __getitem__(self, *args):
        return self.types.__getitem__(*args)



from contextlib import contextmanager

from django.db import connections

from ....core.schema import GraphQLSchema
from ....core.types import Field
from ....plugins import Plugin
from .sql.tracking import unwrap_cursor, wrap_cursor
from .sql.types import DjangoDebugSQL
from .types import DjangoDebug


class WrappedRoot(object):

    def __init__(self, root):
        self._recorded = []
        self._root = root

    def record(self, **log):
        self._recorded.append(DjangoDebugSQL(**log))

    def debug(self):
        return DjangoDebug(sql=self._recorded)


class WrapRoot(object):

    @property
    def _root(self):
        return self._wrapped_root.root

    @_root.setter
    def _root(self, value):
        self._wrapped_root = value

    def resolve_debug(self, args, info):
        return self._wrapped_root.debug()


def debug_objecttype(objecttype):
    return type(
        'Debug{}'.format(objecttype._meta.type_name),
        (WrapRoot, objecttype),
        {'debug': Field(DjangoDebug, name='__debug')})


class DjangoDebugPlugin(Plugin):

    def enable_instrumentation(self, wrapped_root):
        # This is thread-safe because database connections are thread-local.
        for connection in connections.all():
            wrap_cursor(connection, wrapped_root)

    def disable_instrumentation(self):
        for connection in connections.all():
            unwrap_cursor(connection)

    def wrap_schema(self, schema_type):
        query = schema_type._query
        if query:
            class_type = self.schema.objecttype(schema_type.get_query_type())
            assert class_type, 'The query in schema is not constructed with graphene'
            _type = debug_objecttype(class_type)
            self.schema.register(_type, force=True)
            return GraphQLSchema(
                self.schema,
                self.schema.T(_type),
                schema_type.get_mutation_type(),
                schema_type.get_subscription_type()
            )
        return schema_type

    @contextmanager
    def context_execution(self, executor):
        executor['root'] = WrappedRoot(root=executor['root'])
        executor['schema'] = self.wrap_schema(executor['schema'])
        self.enable_instrumentation(executor['root'])
        yield executor
        self.disable_instrumentation()
